%%% The contents of this file are subject to the Erlang Public License,
%%% Version 1.0, (the "License"); you may not use this file except in
%%% compliance with the License. You may obtain a copy of the License at
%%% http://www.erlang.org/license/EPL1_0.txt
%%%
%%% Software distributed under the License is distributed on an "AS IS"
%%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%%% the License for the specific language governing rights and limitations
%%% under the License.
%%%
%%% The Original Code is xmerl-0.19
%%%
%%% The Initial Developer of the Original Code is Ericsson Telecom
%%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%%% Telecom AB. All Rights Reserved.
%%%
%%%
%%%----------------------------------------------------------------------
%%% #0.    BASIC INFORMATION
%%%----------------------------------------------------------------------
%%% File:       xmerl_validate.erl
%%% Author       : Sebastien Becuwe <sebastien.becuwe@cellicium.com>
%%% Description  : XML validation hooks for xmerl
%%% 
%%% Modules used : lists, ets, xmerl_lib
%%% 
%%%----------------------------------------------------------------------

-module(xmerl_validate).

-vsn('0.1').
-date('27-11-02').
-author('sebastien.becuwe@cellicium.com').

-export([validate/2]).

%%%%%%%%%%%%% TODOO
%%%% Atributes data type id , idrefs, FIXED, ...

-include("xmerl.hrl").		% record def, macros

%% +type validate(xmerl_scanner(),xmlElement())->
%%              xmlElment() | {error,tuple()}.
validate(#xmerl_scanner{doctype_name=DTName},#xmlElement{name=Name})
  when DTName=/=Name ->
    {error, {mismatched_root_element,Name,DTName}};
validate(#xmerl_scanner{rules=Rules},XML=#xmlElement{name=Name})->
    catch do_validation(read_rules(Rules,Name),XML,Rules);
validate(_, XML) ->
    {error, {no_xml_element, XML}}.



%% +type validate(rules(),xmlElement())->
%%              {ok,xmlElement()} | {error,tuple()}.
do_validation(undefined,#xmlElement{name=Name}, _Rules) ->
    {error,{unknown_element,Name}};
do_validation(_E, #xmlText{}=XML, _Rules) -> % Nothing to validate
    XML;
do_validation(El_Rule,XML,Rules)->
    case valid_attributes(El_Rule#xmlElement.attributes,
			  XML#xmlElement.attributes) of
	{error,Reason} ->
	    {error,Reason};
	Attr_2->
	    XML_=XML#xmlElement{attributes=Attr_2},
	    case valid_contents(El_Rule#xmlElement.content,
				XML_#xmlElement.content,Rules) of
		{error,Reason}->
		    {error,Reason};
		{error,Reason,N}->
		    {error,Reason,N};
		XMLS ->
		    XML_#xmlElement{content=XMLS}
	    end
    end.


%% +type read_rules(DTD::atom(),Element_Name::atom())->
%%              undefined | xmlElement().
read_rules(_, pcdata) ->
    pcdata;
read_rules(T, Name) ->
    case ets:lookup(T, {elem_def, Name}) of
	[] ->
	    undefined;
	[{_K, V}] ->
	    V
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Attributes Validation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% +deftype attribute_rule() = {Attr_Name::atom(),attribute_type(),
%%                              attribute_priority()}.

%% +type valid_attributes([attribute_rule()],[xmlAttribute()])->
%%              [xmlAttribute()] | {error,attribute_unknow}.
valid_attributes(All_Attr,[#xmlAttribute{}|_T]=Attr)->
    lists:foreach(fun(#xmlAttribute{name=Name})->
			  case is_attribute_exist(Name,All_Attr) of
			      true ->
				  ok;
			      false ->
				  exit({error,{attribute_unknown,Name}})
			  end
		  end,
		  Attr),
    lists:flatten(lists:foldl(fun({Name,DataType,IF},Attr_2)->
				      Attr_2++
					  [valid_attribute(Name,DataType,IF,Attr)] 
			      end,[],All_Attr));
valid_attributes([],[]) ->
    [];
valid_attributes(All_Attr,[]) ->
    lists:flatten(lists:foldl(fun({Name,DataType,IF},Attr_2)->
				      Attr_2++[valid_attribute(Name,
							       DataType,IF,
							       [])] 
			      end,[],All_Attr)).

%%%%  [60]      DefaultDecl::=   
%%%%                              '#REQUIRED' | '#IMPLIED' 
%%%%                            | (('#FIXED' S)? AttValue)
%% +deftype attribute_priority = '#REQUIRED'|'#FIXED'|'#IMPLIED'.

%% +type valid_attribute(Name::atom(),DataType::attribute_value(),
%%                       IF::attribute_priority(),[xmlAttribute()])->
%%         [xmlAttribute()] | exit().
valid_attribute(Name,DataType,IF,List_of_Attributes)->
    case {IF,search_attr(Name,List_of_Attributes)} of
	{'#REQUIRED',no_attribute}->
	    exit({error,{Name,is_required}});
	{'#IMPLIED',no_attribute}->
	    []; %% and no default value
	{Value,no_attribute} when list(Value)->
	    #xmlAttribute{name=Name,value=Value};
	{_,#xmlAttribute{}=Attr}->
	    %% do test data value
	    test_attribute_value(DataType,Attr);
	{Else,XML} ->
	    exit({error,{unknow_attribute_type,Else,XML}})
    end.

search_attr(Name,[#xmlAttribute{name=Name}=H|_T])->
    H;
search_attr(Name,[#xmlAttribute{}|T])-> 
    search_attr(Name,T);
search_attr(_Name,_T) ->
    no_attribute.

is_attribute_exist(Name,[{Name,_,_}|_T])->
    true;
is_attribute_exist(Name,[{_Attr,_,_}|T]) ->
    is_attribute_exist(Name,T);
is_attribute_exist(_Name,[]) ->
    false.

%%%%[54] AttType::=     StringType | TokenizedType | EnumeratedType 
%%%%[55] StringType::=  'CDATA'
%%%%[56] TokenizedType::= 'ID'|'IDREF'| 'IDREFS'|'ENTITY'| 'ENTITIES'
%%%%                     | 'NMTOKEN'| 'NMTOKENS'
%%%%[57] EnumeratedType::= NotationType | Enumeration 
%%%%[58] NotationType::= 'NOTATION' S '(' S? Name (S? '|' S? Name)* S? ')' 
%%%%[59] Enumeration ::= '(' S? Nmtoken (S? '|' S? Nmtoken)* S? ')'

%% +deftype attribute_type()-> 'CDATA' | 'ID'|'IDREF'| 'IDREFS'|'ENTITY'| 
%%                             'ENTITIES'| 'NMTOKEN'| 'NMTOKENS'
%%                             {enumeration,[List_of_value::atom()]}.

%% +type test_attribute_value(attribute_type(),xmlAttribute())->
%%             xmlAttribute()| exit.
%%%% test the constraint validity of Attribute value.
test_attribute_value('CDATA',#xmlAttribute{}=Attr) ->
    Attr;
test_attribute_value('NMTOKEN',#xmlAttribute{name=Name,value=V}=Attr) ->
    lists:foreach(fun (X)->
			  case xmerl_lib:is_namechar(X) of
			      true->
				  ok;
			      false->
				  %%io:format("Warning*** nmtoken,value_incorrect:  ~p~n",[V]),
				  exit({error,{invalid_value_nmtoken,Name,V}})
			  end
		  end, V),
    Attr;
test_attribute_value('NMTOKENS',#xmlAttribute{name=Name,value=V}=Attr) ->
    lists:foreach(fun (X)->
			  case xmerl_lib:is_namechar(X) of
			      true->
				  ok;
			      false when ?whitespace(X)->
				  ok;
			      false ->
				  exit({error,{invalid_value_nmtokens,Name,V}})
			  end
		  end, V),
    Attr;
test_attribute_value({enumeration,L},#xmlAttribute{value=Value}=Attr) ->
    case lists:member(list_to_atom(Value),L) of
	true->
	    Attr;
	false ->
	    exit({error,{attribute_value_unknow,Value,{list,L}}})
    end;
test_attribute_value(Rule,Attr) ->
%    io:format("Attr Value*****~nRule~p~nValue~p~n",[Rule,Attr]),
    Attr.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Contents Validation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%Element-content Models
%%%%[47] children::= (choice | seq) ('?' | '*' | '+')?
%%%%[48] cp::=       (Name | choice | seq) ('?' | '*' | '+')?
%%%%[49] choice::=  '(' S? cp ( S? '|' S? cp )+ S? ')'
%%%%[50] seq::=     '(' S? cp ( S? ',' S? cp )* S? ')'
%%%%[51] Mixed::=   '(' S? '#PCDATA' (S? '|' S? Name)* S? ')*' 
%%%%              | '(' S? '#PCDATA' S? ')' 


%% +type valid_contents([rule()],[xmlElement()])->
%%              [xmlElement() | {error,???}.
valid_contents(Rule,XMLS,Rules)->
    case parse(Rule,XMLS,Rules) of
	{XML_N,[]}->
	    lists:flatten(XML_N);
	{_,[#xmlElement{name=Name}|_T]} ->
	    exit({error,{element,Name,isnt_comprise_in_the_rule}});
	{_,[#xmlText{}|_T]} ->
	    exit({error,{element,text,isnt_comprise_in_the_rule}});
	{error,Reason} ->
	    {error,Reason};
	{error,Reason,N} ->
	    {error,Reason,N}
    end.

parse({'*',SubRule},XMLS,Rules)->
    star(SubRule,XMLS,Rules); 
parse({'+',SubRule},XMLS,Rules) ->
    plus(SubRule,XMLS,Rules);
parse({choice,CHOICE},XMLS,Rules)->
    choice(CHOICE,XMLS,Rules);
parse(empty,[],_Rules) ->
    {[],[]};
parse({'?',SubRule},XMLS,Rules)->
    question(SubRule,XMLS,Rules);
parse({seq,List},XMLS,Rules) ->
    seq(List,XMLS,Rules);
parse(El_Name,[#xmlElement{name=El_Name}=XML|T],Rules) when atom(El_Name)->
    case do_validation(read_rules(Rules,el_name(XML)),XML,Rules) of
	{error,R} ->
	    {error,R};
	{error,R,N}->
	    {error,R,N};
	XML_->
	    {[XML_],T}
    end;
parse(El_Name,[#xmlElement{name=Name}|_T]=S,_Rules) when atom(El_Name)->
    {error,
     {element_seq_not_conform,{wait,El_Name},{is,Name}},
     {{next,S},{act,[]}} };
parse('#PCDATA',XML,_Rules)->
    %%% PCDATA it is 0 , 1 or more #xmlText{}.
    parse_pcdata(XML);
parse(El_Name,[#xmlText{}|_T]=S,_Rules)->
    {error,
     {text_in_place_of,El_Name},
     {{next,S},{act,[]}}};
parse([],_,_) ->
    {error,no_rule};
parse(Rule,[],_) ->
    {error,{no_xml_element,Rule}}.


choice([CH|CHS],[_XML|_T]=XMLS,Rules)->
    case parse(CH,XMLS,Rules) of
	{error,_R} ->
	    choice(CHS,XMLS,Rules);
	{error,_R,_N} ->
	    choice(CHS,XMLS,Rules);
	{Tree,XMLS2}->
	    {[Tree],XMLS2}
    end;
choice([],XMLS,_)->
    {error,
     element_unauthorize_in_choice,
     {{next,XMLS},{act,[]}}}.

plus(Rule,XMLS,Rules) ->
    %% 1 or more
    case parse(Rule,XMLS,Rules) of
	{error, Reason,_XML} ->
	    {error, Reason};
	{error, X} ->
	    {error, X};
	{Tree, XMLS1} ->
	    case star(Rule, XMLS1,Rules) of
		{error, X}->
		    {error,X};
		{[], _} ->
		    {[Tree], XMLS1};
		{Tree_1, XMLS2} ->
		    {[Tree]++Tree_1, XMLS2}
	    end
    end.

star(Rule, XMLS, Rules) ->
    %% zero or more
    case star2(Rule,XMLS, Rules,[]) of
	{error, _} ->
	    {[], XMLS};
	{Tree, Toks1} ->
	    {Tree, Toks1}
    end.

star2(_Rule,XML,_Rules,Tree) when length(XML)==0->
    {[Tree],[]};
star2(Rule,XMLS,Rules,Tree) ->
    case parse(Rule,XMLS,Rules) of
	{error, _E, {{next,N},{act,A}}}->
	    %%io:format("Error~p~n",[_E]),
	    {Tree++A,N};
	{error, _E}->
	    %%io:format("Error~p~n",[_E]),
	    {[Tree],[]};
	{Tree1,XMLS_1}->
	    star2(Rule,XMLS_1,Rules,Tree++[Tree1])
    end.

question(Rule, Toks,Rules) ->
    %% 0 or 1
    case parse(Rule, Toks,Rules) of
	{error, _E, _Next}->
	    {[],Toks};
	{error, _E} ->
	    {[], Toks};
	{T,Toks1} -> 
	    {T, Toks1}
    end.

seq(H,Toks,Rules)->
    case seq2(H,Toks,Rules,[]) of
	{error,E}->
	    {error,E};
	{error,R,N}->
	    {error,R,N};
	{Tree,Toks2}->
	    {Tree,Toks2}
    end.

seq2([],[],_,Tree)->
    {Tree,[]};
seq2([],[#xmlElement{name=Name}|_T]=XMLS,_,Tree)->
    {error,
     {sequence_finish,Name,isnt_in_the_right_place},
     {{next,XMLS},{act,Tree}}};
seq2([],[#xmlText{}]=XML,_,Tree)->
    {error,
     sequence_finish,
     {{next,XML},{act,Tree}}};
seq2([H|T],Toks,Rules,Tree) ->
    case parse (H,Toks,Rules) of
	{error,Reason,_XML}->
	    {error,Reason};
	{error,E}->
	    {error,E};
	{[],Toks2}->
	    seq2(T,Toks2,Rules,Tree);
	{Tree1,Toks2} when list(Tree1)->
	    seq2(T,Toks2,Rules,Tree++Tree1);
	{Tree1,Toks2}->
	    seq2(T,Toks2,Rules,Tree++[Tree1])
    end.
el_name(#xmlElement{name=Name})->
    Name;
el_name(#xmlText{}) ->
    pcdata.

parse_pcdata([#xmlText{}=H|T])->
    parse_pcdata(T,[H]);
parse_pcdata(H) ->
    {[],H}.

parse_pcdata([#xmlText{}=H|T],Acc)->
    parse_pcdata(T,Acc++[H]);
parse_pcdata(H,Acc) ->
    {Acc,H}.
